
#include "Boiler.h"
#include "Fluid.h"
#include "Resources.h"
#include "../utils/DXHelpers.h"
#include "../Vertices.h"
#include "../engine/Camera.h"
#include <cassert>
#include <cstdlib>
#include <cmath>

const int MAX_BUBBLES = 2000;


CBoiler::CBoiler( CResources& resources, CCamera& camera )
:	mResources( &resources ),
	mCamera( &camera ),
	mShowingVelocity( false ),
	mDrawingWireframe( false ),
	mWaterSurface( resources ),
	mBurbulator( resources ),
	mTime( 0.0f ),
	mVBWater( NULL ),
	mIBBubbles( NULL ),
	mVBBubbles( NULL )
{
}

CBoiler::~CBoiler()
{
}


HRESULT CBoiler::onCreateDevice( IDirect3DDevice8* device )
{
	CDXObject::onCreateDevice( device );
	return S_OK;
}

HRESULT CBoiler::onLostDevice()
{
	CDXObject::onLostDevice();

	dx::assertRelease( mVBWater );
	dx::assertRelease( mIBBubbles );
	dx::assertRelease( mVBBubbles );
	
	return S_OK;
}

HRESULT CBoiler::onResetDevice()
{
	CDXObject::onResetDevice();

	// VB lines
	assert( !mVBWater );
	int maxVerts = mResources->mFluid->getTotalSize() * 2;
	getDevice().CreateVertexBuffer(
		maxVerts * sizeof(SVertexXyzDiffuse), D3DUSAGE_WRITEONLY | D3DUSAGE_DYNAMIC,
		FVF_XYZ_DIFFUSE, D3DPOOL_DEFAULT,
		&mVBWater
	);
	assert( mVBWater );

	// VB/IB bubbles
	createBubbleBuffers();
	
	return S_OK;
}

HRESULT CBoiler::onDestroyDevice()
{
	CDXObject::onDestroyDevice();
	
	return S_OK;
}

void CBoiler::createBubbleBuffers()
{
	assert( !mIBBubbles );
	getDevice().CreateIndexBuffer(
		MAX_BUBBLES * 6 * sizeof(short), D3DUSAGE_WRITEONLY,
		D3DFMT_INDEX16, D3DPOOL_DEFAULT,
		&mIBBubbles
		);
	assert( mIBBubbles );
	short* ib = NULL;
	mIBBubbles->Lock( 0, 0, (BYTE**)&ib, 0 );
	assert( ib );
	for( int i = 0; i < MAX_BUBBLES; ++i ) {
		short base = i * 4;
		ib[0] = base;
		ib[1] = base+1;
		ib[2] = base+2;
		ib[3] = base;
		ib[4] = base+2;
		ib[5] = base+3;
		ib += 6;
	}
	mIBBubbles->Unlock();
	
	assert( !mVBBubbles );
	getDevice().CreateVertexBuffer(
		MAX_BUBBLES * 4 * sizeof(SVertexXyzTex1), D3DUSAGE_WRITEONLY | D3DUSAGE_DYNAMIC,
		FVF_XYZ_TEX1, D3DPOOL_DEFAULT,
		&mVBBubbles
		);
	assert( mVBBubbles );
}

void CBoiler::fillParticle( SVertexXyzTex1* vb, const D3DXVECTOR3& pos, float size, const D3DXMATRIX& viewMat )
{
	static D3DXVECTOR3 cornerUL( -1.0f, 1.0f,  0.0f );
	static D3DXVECTOR3 cornerUR(  1.0f, 1.0f,  0.0f );
	static D3DXVECTOR3 cornerDL( -1.0f, -1.0f, 0.0f );
	static D3DXVECTOR3 cornerDR(  1.0f, -1.0f, 0.0f );

	// 4 particle corners
	D3DXVECTOR3 v1;
	D3DXVec3TransformCoord( &v1, &(cornerUL * size), &viewMat );
	vb->p = pos + v1;
	vb->tu = 0.0f; vb->tv = 0.0f;
	++vb;
	D3DXVECTOR3 v2;
	D3DXVec3TransformCoord( &v2, &(cornerUR * size), &viewMat );
	vb->p = pos + v2;
	vb->tu = 1.0f; vb->tv = 0.0f;
	++vb;
	D3DXVECTOR3 v3;
	D3DXVec3TransformCoord( &v3, &(cornerDR * size), &viewMat );
	vb->p = pos + v3;
	vb->tu = 1.0f; vb->tv = 1.0f;
	++vb;
	D3DXVECTOR3 v4;
	D3DXVec3TransformCoord( &v4, &(cornerDL * size), &viewMat );
	vb->p = pos + v4;
	vb->tu = 0.0f; vb->tv = 1.0f;
	++vb;
}

void CBoiler::render( float lastFrameTime )
{
	mTime += lastFrameTime;

	HRESULT hr;

	CFluid& fluid = *mResources->mFluid;
	int size = fluid.getSize();
	int nPoints = fluid.getTotalSize();
	float scale = (size/2) / R_FLUID_RADIUS;
	float step = R_FLUID_RADIUS / (size/2);
	
	UINT passes, p;

	// ----------------------------------------------------------
	//  velocity

	if( mShowingVelocity )
		renderVelocity( lastFrameTime );

	// ----------------------------------------------------------
	//  bowl

	getDevice().SetRenderState( D3DRS_FILLMODE, mDrawingWireframe ? D3DFILL_WIREFRAME : D3DFILL_SOLID );

	getDevice().SetMaterial( &mResources->mMaterialBowl );
	mResources->mEfxBaseTexEnvi->Begin( &passes, 0 );
	for( p = 0; p < passes; ++p ) {
		mResources->mEfxBaseTexEnvi->Pass( p );
		mResources->mMeshBowl->DrawSubset( 0 );
	}
	mResources->mEfxBaseTexEnvi->End();

	getDevice().SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	
	// ----------------------------------------------------------
	//  bubbles
	
	if( !mBurbulator.getBubbles().empty() ) {
		D3DXMATRIX viewMatrix = mCamera->getCameraMatrix();
		viewMatrix._41 = viewMatrix._42 = viewMatrix._43 = 0.0f;

		SVertexXyzTex1* vbb = NULL;
		int count = mBurbulator.getBubbles().size();
		if( count > MAX_BUBBLES )
			count = MAX_BUBBLES;
		hr = mVBBubbles->Lock( 0, count * 4 * sizeof(SVertexXyzTex1), (BYTE**)&vbb, D3DLOCK_DISCARD );
		assert( vbb );
		CBurbulator::TBubblePool::const_iterator it;
		int i = 0;
		for( it = mBurbulator.getBubbles().begin(); it != mBurbulator.getBubbles().end(); ++it ) {
			if( i >= MAX_BUBBLES )
				break;
			const CBurbulator::SBubble& b = *it;
			fillParticle( vbb, b.mPosition, b.mSize, viewMatrix );
			vbb += 4;
			++i;
		}
		hr = mVBBubbles->Unlock();
		hr = getDevice().SetIndices( mIBBubbles, 0 );
		hr = getDevice().SetStreamSource( 0, mVBBubbles, sizeof(SVertexXyzTex1) );
		mResources->mEfxParticle->Begin( &passes, 0 );
		for( p = 0; p < passes; ++p ) {
			mResources->mEfxParticle->Pass( p );
			hr = getDevice().DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, count * 4, 0, count * 2 );
		}
		mResources->mEfxParticle->End();
	}

	// ----------------------------------------------------------
	//  water surface

	getDevice().SetRenderState( D3DRS_FILLMODE, mDrawingWireframe ? D3DFILL_WIREFRAME : D3DFILL_SOLID );

	getDevice().SetMaterial( &mResources->mMaterialWater );
	mResources->mEfxBlendEnvi->Begin( &passes, 0 );
	for( p = 0; p < passes; ++p ) {
		mResources->mEfxBlendEnvi->Pass( p );
		mResources->mMeshWaterDynamic->DrawSubset( 0 );
	}
	mResources->mEfxBlendEnvi->End();
	
	getDevice().SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );

	//
	// update objects

	fluid.setDeltaTime( lastFrameTime * 5.0f );
	fluid.solveVelocity();
	fluid.solveTemperature();
	fluid.resetVelocity0();

	mWaterSurface.updateWaterMesh( lastFrameTime );
	mBurbulator.update( lastFrameTime * 5.0f, mWaterSurface );
}

D3DCOLOR CBoiler::getColorFromTemperature( float temp )
{
	int r = (int)(temp * 255.0f / 100.0f);
	int g = temp <= 50.0f ? 0 : (int)((temp-50.0f) * 127.0f / 50.0f) * 2;
	int b = 255 - (int)(temp * 255.0f / 100.0f);
	return 0x80000000 | ((r&255)<<16) | (g<<8) | (b&255);
}

void CBoiler::renderVelocity( float lastFrameTime )
{
	HRESULT hr;

	CFluid& fluid = *mResources->mFluid;
	int size = fluid.getSize();
	int nPoints = fluid.getTotalSize();
	float scale = (size/2) / R_FLUID_RADIUS;
	float step = R_FLUID_RADIUS / (size/2);
	
	//
	// debug water speeds

	SVertexXyzDiffuse* vb = NULL;
	hr = mVBWater->Lock( 0, nPoints * 2 * sizeof(SVertexXyzDiffuse), (BYTE**)&vb, D3DLOCK_DISCARD );
	assert( vb );
	int realPoints = 0;
	const float* velX = fluid.getVelocityX();
	const float* velY = fluid.getVelocityY();
	const float* velZ = fluid.getVelocityZ();
	const float* temp = fluid.getTemperature();
	float ini = -R_FLUID_RADIUS + step/2;
	float z = step/2;
	for( int k = 0; k < size; ++k, z += step ) {
		float y = ini;
		for( int j = 0; j < size; ++j, y += step ) {
			float x = ini;
			for( int i = 0; i < size; ++i, x += step ) {
				if( x*x+y*y <= R_FLUID_RADIUS2*1.7f && z <= R_FLUID_HEIGHT ) {
					D3DXVECTOR3 pos( x, y, z );
					D3DCOLOR col = getColorFromTemperature( *temp );
					vb->p = pos;
					vb->diffuse = col;
					++vb;
					D3DXVECTOR3 len( *velX, *velY, *velZ );
					vb->p = pos + len * size * step;
					vb->diffuse = col;
					++vb;
					++realPoints;
				}
				++velX;
				++velY;
				++velZ;
				++temp;
			}
		}
	}
	hr = mVBWater->Unlock();

	hr = getDevice().SetStreamSource( 0, mVBWater, sizeof(SVertexXyzDiffuse) );

	UINT passes;
	UINT p;
	mResources->mEfxWater->Begin( &passes, 0 );
	for( p = 0; p < passes; ++p ) {
		mResources->mEfxWater->Pass( p );
		hr = getDevice().DrawPrimitive( D3DPT_LINELIST, 0, realPoints );
	}
	mResources->mEfxWater->End();
}
